from .entries import SystemProps, load_system  # noqa: F401
from .globals.libs import (  # noqa: F401
    extract_cross_layer_props,
)
from .globals.logging import (  # noqa: F401
    composite_logger,
    console_log_full,
    console_log_json,
    console_log_simple,
    log_tcp,
    standard_logger,
)
from .libs import (  # noqa: F401
    combine_cross_layer_props,
    create_error_object,
    get_layers_unavailable,
    get_log_level_name,
    get_log_level_number,
    is_config,
    is_error_object,
    normalize_cross_layer_props,
    validate_config,
)
from .models.libs import get_model_definition, model  # noqa: F401
from .models.protocols import (  # noqa: F401
    BackendProtocol,
    DatastoreValueType,
    DatesAfterQuery,
    DatesBeforeQuery,
    EqualitySymbol,
    InLayersModel,
    InLayersModelInstance,
    ModelDefinition,
    ModelSearch,
    ModelSearchResult,
    ModelServices,
    PrimaryKeyType,
    PropertyOptions,
    PropertyQuery,
    Query,
    QueryTokens,
    SortOrder,
    SortStatement,
)
from .models.query import (  # noqa: F401
    query_builder,
)
from .protocols import (  # noqa: F401
    AppLayer,
    CommonContext,
    Config,
    CoreConfig,
    CoreLoggingConfig,
    CoreNamespace,
    CrossLayerProps,
    Domain,
    ErrorObject,
    FeaturesContext,
    FunctionLogger,
    GlobalsServicesProps,
    HighLevelLogger,
    LayerContext,
    LayerDescription,
    LayerLogger,
    LogFormat,
    Logger,
    LogId,
    LogLevel,
    LogLevelNames,
    LogMessage,
    LogMethod,
    RootLogger,
)



from __future__ import annotations

import base64
import json
import logging
import uuid
from collections.abc import Callable, Mapping, Sequence
from dataclasses import asdict, is_dataclass
from datetime import UTC, datetime
from enum import Enum
from typing import Any

import httpx
from pydantic import RootModel

from ..libs import (
    combine_cross_layer_props,
    create_error_object,
    get_log_level_number,
)
from ..protocols import (
    AppLogger,
    CommonContext,
    CommonLayerName,
    CrossLayerProps,
    ErrorObject,
    HighLevelLogger,
    LayerLogger,
    LogFormat,
    Logger,
    LogId,
    LogLevel,
    LogLevelNames,
    LogMessage,
    LogMethod,
    RootLogger,
)
from .libs import (
    cap_for_logging,
    combine_logging_props,
    default_get_function_wrap_log_level,
    extract_cross_layer_props,
)

MAX_LOGGING_ATTEMPTS = 5
DEFAULT_MAX_LOG_SIZE_IN_CHARACTERS = 50000


def _handle_special_types(
    obj: Any, max_depth: int, _depth: int, _seen: set[int]
) -> tuple[bool, Any]:
    """Handle Enum, datetime, and bytes types. Returns (handled, value)."""
    if isinstance(obj, Enum):
        return True, _to_jsonable(
            obj.value, max_depth=max_depth, _depth=_depth + 1, _seen=_seen
        )
    if isinstance(obj, datetime):
        return True, obj.isoformat()
    if isinstance(obj, (bytes, bytearray, memoryview)):
        b = bytes(obj)
        try:
            return True, b.decode("utf-8")
        except Exception:
            return True, {"$bytes_b64": base64.b64encode(b).decode("ascii")}
    return False, None


def _handle_structured_types(
    obj: Any, max_depth: int, _depth: int, _seen: set[int]
) -> tuple[bool, Any]:
    """Handle Pydantic models, dataclasses, and exceptions. Returns (handled, value)."""
    if hasattr(obj, "model_dump") and callable(obj.model_dump):
        try:
            dumped = obj.model_dump(mode="json")  # type: ignore[call-arg]
        except Exception:
            dumped = obj.model_dump()  # type: ignore[call-arg]
        return True, _to_jsonable(
            dumped, max_depth=max_depth, _depth=_depth + 1, _seen=_seen
        )
    if is_dataclass(obj):
        return True, _to_jsonable(
            asdict(obj), max_depth=max_depth, _depth=_depth + 1, _seen=_seen
        )
    if isinstance(obj, BaseException):
        return True, {
            "type": type(obj).__name__,
            "message": str(obj),
            "args": _to_jsonable(
                list(getattr(obj, "args", ()) or ()),
                max_depth=max_depth,
                _depth=_depth + 1,
                _seen=_seen,
            ),
        }
    return False, None


def _handle_collections(
    obj: Any, max_depth: int, _depth: int, _seen: set[int]
) -> tuple[bool, Any]:
    """Handle mappings and sequences. Returns (handled, value)."""
    if isinstance(obj, Mapping):
        out: dict[str, Any] = {}
        for k, v in obj.items():
            out[str(k)] = _to_jsonable(
                v, max_depth=max_depth, _depth=_depth + 1, _seen=_seen
            )
        return True, out
    if isinstance(obj, (list, tuple, set, frozenset)):
        return True, [
            _to_jsonable(x, max_depth=max_depth, _depth=_depth + 1, _seen=_seen)
            for x in list(obj)
        ]
    return False, None


def _to_jsonable(
    obj: Any,
    *,
    max_depth: int = 6,
    _depth: int = 0,
    _seen: set[int] | None = None,
) -> Any:
    """
    Convert arbitrary objects into JSON-serializable structures.

    This is intentionally lossy: anything unknown becomes a string.
    """
    if _seen is None:
        _seen = set()
    if _depth > max_depth:
        return "[MaxDepth]"

    # Fast-path JSON primitives
    if obj is None or isinstance(obj, (bool, int, float, str)):
        return obj

    # Avoid cycles
    oid = id(obj)
    if oid in _seen:
        return "[Circular]"
    _seen.add(oid)

    handlers = [
        _handle_special_types,
        _handle_structured_types,
        _handle_collections,
    ]
    try:
        for handler in handlers:
            handled, value = handler(obj, max_depth, _depth, _seen)
            if handled:
                return value
        # As a last resort, prefer stringifying (stable + safe)
        return str(obj)
    finally:
        # allow the same object to appear elsewhere without being treated as a cycle
        _seen.discard(oid)


def _setup_python_logging(level: int) -> None:
    root = logging.getLogger()
    root.setLevel(level)
    fmt = logging.Formatter("%(message)s")
    for h in root.handlers:
        h.setFormatter(fmt)

    if not root.handlers:
        h = logging.StreamHandler()
        h.setLevel(level)
        h.setFormatter(logging.Formatter("%(message)s"))
        root.addHandler(h)


def _to_std_level(name: LogLevelNames | None) -> int:
    if name == LogLevelNames.error:
        return logging.ERROR
    if name == LogLevelNames.warn:
        return logging.WARNING
    if name == LogLevelNames.info:
        return logging.INFO
    # Map both trace and debug to DEBUG for std logging
    return logging.DEBUG


def console_log_simple(log_message: LogMessage) -> None:
    splitted = (log_message.get("logger") or "root").split(":")
    function_name = splitted[-1] if splitted else "root"
    msg = f"{log_message['datetime'].isoformat()}: {function_name} {log_message['message']}"
    logging.log(_to_std_level(log_message.get("log_level")), msg)


def _combine_ids(ids: list[LogId]) -> str:
    parts: list[str] = []
    for obj in ids:
        parts.append(";".join([f"{k}:{v}" for k, v in obj.items()]))
    return ";".join(parts)


def console_log_full(log_message: LogMessage) -> None:
    ids = log_message.get("ids")
    level_obj = log_message.get("log_level")
    level_str = (
        level_obj.value
        if hasattr(level_obj, "value")
        else (str(level_obj) if level_obj is not None else "")
    )
    if ids:
        msg = (
            f"{log_message['datetime'].isoformat()} {log_message.get('environment')} {level_str} {log_message['id']} "
            f"[{log_message.get('logger')}] {{{_combine_ids(ids)}}} {log_message['message']}"
        )
    else:
        msg = (
            f"{log_message['datetime'].isoformat()} {log_message.get('environment')} {level_str} "
            f"[{log_message.get('logger')}] {log_message['message']}"
        )
    logging.log(_to_std_level(log_message.get("log_level")), msg)


def console_log_json(log_message: LogMessage) -> None:
    base = {
        "id": log_message.get("id"),
        "datetime": log_message["datetime"].isoformat(),
        "log_level": log_message.get("log_level"),
        "logger": log_message.get("logger"),
        "message": log_message.get("message"),
    }
    rest = dict(log_message)
    for k in ["id", "datetime", "log_level", "logger", "message"]:
        rest.pop(k, None)
    msg = json.dumps({**base, **rest}, default=str)
    logging.log(_to_std_level(log_message.get("log_level")), msg)


def log_tcp(context: CommonContext) -> Callable[[LogMessage], Any]:
    tcp_options = context.config.in_layers_core.logging.tcp_logging_options
    if not tcp_options:
        raise ValueError("Must include tcp_logging_options when using a tcp logger")
    url = tcp_options.url
    headers = tcp_options.headers or {}
    client = httpx.Client(base_url=url, headers=headers)

    def _send(log_message: LogMessage) -> Any:
        success: bool | None = None
        for _ in range(MAX_LOGGING_ATTEMPTS):
            try:
                client.post("", json=log_message)
                success = True
                break
            except Exception:
                logging.exception("Logging error")
                success = False
        return success

    return _send


def _should_ignore(config_level: LogLevelNames, message_level: LogLevelNames) -> bool:
    a = get_log_level_number(config_level)
    if a == LogLevel.SILENT.value:
        return True
    b = get_log_level_number(message_level)
    return a > b


def _get_log_methods_from_format(
    log_format: LogFormat | list[LogFormat],
) -> list[LogMethod]:
    if isinstance(log_format, list):
        result: list[LogMethod] = []
        for lf in log_format:
            result.extend(_get_log_methods_from_format(lf))
        return result
    if log_format == LogFormat.custom:
        raise ValueError(
            "This should never be here. custom_logger should override this"
        )
    if log_format == LogFormat.json:
        return [lambda _ctx: console_log_json]
    if log_format == LogFormat.simple:
        return [lambda _ctx: console_log_simple]
    if log_format == LogFormat.full:
        return [lambda _ctx: console_log_full]
    if log_format == LogFormat.tcp:
        return [log_tcp]
    raise ValueError(f"LogFormat {log_format} is not supported")


class _HL(HighLevelLogger):  # type: ignore[misc]
    def __init__(self, context: CommonContext, sub: Logger):
        self.context = context
        self.sub = sub

    def get_app_logger(self, app_name: str) -> AppLogger:
        return _app_logger(self.context, self.sub, app_name)

    # Logger methods forwarded
    def trace(self, *a, **k):
        return self.sub.trace(*a, **k)

    def debug(self, *a, **k):
        return self.sub.debug(*a, **k)

    def info(self, *a, **k):
        return self.sub.info(*a, **k)

    def warn(self, *a, **k):
        return self.sub.warn(*a, **k)

    def error(self, *a, **k):
        return self.sub.error(*a, **k)

    def apply_data(self, *a, **k):
        return self.sub.apply_data(*a, **k)

    def get_id_logger(self, *a, **k):
        return self.sub.get_id_logger(*a, **k)

    def get_sub_logger(self, *a, **k):
        return self.sub.get_sub_logger(*a, **k)

    def get_ids(self):
        return self.sub.get_ids()


def composite_logger(log_methods: Sequence[LogMethod]) -> RootLogger:
    def get_logger(
        context: CommonContext, props: Mapping[str, Any] | None = None
    ) -> HighLevelLogger:
        ids = _get_ids_with_runtime(context["constants"]["runtime_id"], props)
        sub = _sub_logger(
            context,
            list(log_methods),
            {"names": [], "ids": ids, "data": dict(props or {}).get("data", {})},
        )
        return _HL(context, sub)

    class _Root(RootLogger):  # type: ignore[misc]
        def get_logger(
            self, context: CommonContext, props: Mapping[str, Any] | None = None
        ) -> HighLevelLogger:
            return get_logger(context, props)

    return _Root()


def standard_logger() -> RootLogger:
    class _Root(RootLogger):  # type: ignore[misc]
        def get_logger(
            self, context: CommonContext, props: Mapping[str, Any] | None = None
        ) -> HighLevelLogger:
            _setup_python_logging(
                _to_std_level(context.config.in_layers_core.logging.log_level)
            )
            logging_cfg = context.config.in_layers_core.logging
            custom = logging_cfg.get("custom_logger", None)
            if custom:
                ids = _get_ids_with_runtime(context.constants.runtime_id, props)
                return custom.get_logger(context, {**(props or {}), "ids": ids})
            methods = _get_log_methods_from_format(logging_cfg.log_format)
            return composite_logger(methods).get_logger(context, props)

    return _Root()


class _Logger(Logger):  # type: ignore[misc]
    def __init__(
        self,
        context: CommonContext,
        log_methods: list[LogMethod],
        props: dict[str, Any],
    ):
        self.context = context
        self.props = props
        self.config_level: LogLevelNames = (
            context.config.in_layers_core.logging.log_level
        )
        self.bound_methods: list[Callable[[LogMessage], Any]] = [
            m(context) for m in log_methods
        ]
        self.log_methods = log_methods

    def _do_log(self, message_level: LogLevelNames):
        def _f(
            message: str,
            data_or_error: Mapping[str, Any] | ErrorObject | None = None,
            *,
            ignore_size_limit: bool = False,
        ) -> Any:
            if _should_ignore(self.config_level, message_level):
                return None
            is_error_obj = isinstance(data_or_error, ErrorObject)
            is_error_like = isinstance(data_or_error, Mapping) and "error" in (
                data_or_error or {}
            )
            data = {}
            if is_error_obj:
                data = RootModel(data_or_error).model_dump()
            elif is_error_like:
                data = dict(data_or_error or {})
            else:
                data = dict(data_or_error or {})
            # Ensure that logged payloads are JSON-serializable so consumers
            # (e.g. tcp logger / external pipelines) don't need to normalize.
            jsonable_data = _to_jsonable(data)
            error_jsonable = (
                jsonable_data.get("error")
                if isinstance(jsonable_data, Mapping)
                and (is_error_like or is_error_obj)
                else None
            )
            the_data = (
                jsonable_data
                if ignore_size_limit
                else cap_for_logging(
                    jsonable_data,
                    self.context.config.in_layers_core.logging.get(
                        "max_log_size_in_characters", DEFAULT_MAX_LOG_SIZE_IN_CHARACTERS
                    )
                    or 50000,
                )
            )
            if (is_error_like or is_error_obj) and error_jsonable is not None:
                # Preserve `error` even if truncation dropped it.
                try:
                    the_data = dict(the_data or {})
                    the_data["error"] = error_jsonable
                except Exception:
                    the_data = {"error": error_jsonable}
            log_message: LogMessage = {
                "id": str(uuid.uuid4()),
                "environment": self.context.constants["environment"],
                "datetime": datetime.now(tz=UTC),
                "log_level": message_level,
                "message": message,
                "ids": self.props.get("ids"),
                "logger": ":".join(self.props.get("names", [])),
                **the_data,
            }  # type: ignore[typeddict-item]
            [bm(log_message) for bm in self.bound_methods]
            return None

        return _f

    def get_ids(self) -> list[LogId]:
        return list(self.props.get("ids") or [])

    def debug(self, *a, **k):
        return self._do_log(LogLevelNames.debug)(*a, **k)

    def info(self, *a, **k):
        return self._do_log(LogLevelNames.info)(*a, **k)

    def warn(self, *a, **k):
        return self._do_log(LogLevelNames.warn)(*a, **k)

    def trace(self, *a, **k):
        return self._do_log(LogLevelNames.debug)(*a, **k)

    def error(self, *a, **k):
        return self._do_log(LogLevelNames.error)(*a, **k)

    def get_sub_logger(self, name: str) -> Logger:
        return _sub_logger(
            self.context,
            self.log_methods,
            {**self.props, "names": [*self.props.get("names", []), name]},
        )

    def get_id_logger(
        self, name: str, log_id_or_key: LogId | str, id: str | None = None
    ) -> Logger:
        if not isinstance(log_id_or_key, Mapping) and not id:
            raise ValueError("Need value if providing a key")
        log_id: LogId = (
            log_id_or_key
            if isinstance(log_id_or_key, Mapping)
            else {str(log_id_or_key): str(id or "")}
        )
        ids = [*self.props.get("ids", []), log_id]
        return _sub_logger(
            self.context,
            self.log_methods,
            {**self.props, "names": [*self.props.get("names", []), name], "ids": ids},
        )

    def apply_data(self, data: Mapping[str, Any]) -> Logger:
        merged = dict(self.props)
        merged.update(data)
        if "ids" not in data:
            merged["ids"] = self.props.get("ids")
        return _sub_logger(self.context, self.log_methods, merged)


def _sub_logger(
    context: CommonContext,
    log_methods: list[LogMethod],
    props: dict[str, Any],
) -> Logger:
    return _Logger(context, log_methods, props)


class _AppLoggerImpl(AppLogger):  # type: ignore[misc]
    def __init__(self, ctx: CommonContext, base_logger: Logger):
        self._ctx = ctx
        self._base = base_logger

    def get_layer_logger(
        self,
        layer_name: CommonLayerName | str,
        cross_layer_props: CrossLayerProps | None = None,
    ) -> LayerLogger:
        return _layer_logger(self._ctx, self._base, str(layer_name), cross_layer_props)

    def trace(self, *a, **k):
        return self._base.trace(*a, **k)

    def debug(self, *a, **k):
        return self._base.debug(*a, **k)

    def info(self, *a, **k):
        return self._base.info(*a, **k)

    def warn(self, *a, **k):
        return self._base.warn(*a, **k)

    def error(self, *a, **k):
        return self._base.error(*a, **k)

    def apply_data(self, *a, **k):
        return self._base.apply_data(*a, **k)

    def get_id_logger(self, *a, **k):
        return self._base.get_id_logger(*a, **k)

    def get_sub_logger(self, *a, **k):
        return self._base.get_sub_logger(*a, **k)

    def get_ids(self):
        return self._base.get_ids()


def _app_logger(context: CommonContext, sub_logger: Logger, app_name: str) -> AppLogger:
    the_logger = sub_logger.get_sub_logger(app_name).apply_data({"app": app_name})
    return _AppLoggerImpl(context, the_logger)


class _LayerLoggerImpl(LayerLogger):  # type: ignore[misc]
    def __init__(self, ctx: CommonContext, lname: str, base: Logger):
        self._ctx = ctx
        self._layer = lname
        self._base = base

    def get_function_logger(
        self, name: str, cross_layer_props: CrossLayerProps | None = None
    ) -> Logger:
        func_logger = self._base.get_id_logger(
            name, "function_call_id", str(uuid.uuid4())
        ).apply_data({"function": name})
        combined = combine_cross_layer_props(
            {"logging": {"ids": func_logger.get_ids()}},
            cross_layer_props or {"logging": {"ids": []}},
        )
        return func_logger.apply_data(combined["logging"])

    def get_inner_logger(
        self, function_name: str, cross_layer_props: CrossLayerProps | None = None
    ) -> Logger:
        func_logger = self._base.get_sub_logger(function_name).apply_data(
            {"function": function_name}
        )
        combined = combine_cross_layer_props(
            {"logging": {"ids": func_logger.get_ids()}},
            cross_layer_props or {"logging": {"ids": []}},
        )
        return func_logger.apply_data(combined["logging"])

    def _log_wrap(
        self, function_name: str, func: Callable[..., Any]
    ) -> Callable[..., Any]:
        layer = self._layer

        def _wrapped(*args: Any, **kwargs: Any) -> Any:
            # Extract cross-layer props from positional or keyword args
            args_no_cross, kwargs_no_cross, cross_layer_props = (
                extract_cross_layer_props(list(args), dict(kwargs))
            )
            flog = self.get_function_logger(function_name, cross_layer_props)
            level = _get_wrap_level(self._ctx, layer, function_name)
            getattr(flog, level)(f"Executing {layer} function", {"args": args_no_cross})
            try:
                # Always provide the combined logging ids to the inner function wrapper
                combined = combine_cross_layer_props(
                    {"logging": {"ids": flog.get_ids()}}, cross_layer_props or {}
                )
                # Pass combined cross to inner wrapper via keyword; it will decide final forwarding
                result = func(flog, *args_no_cross, **kwargs_no_cross, cross_layer_props=combined)  # type: ignore[arg-type]
                getattr(flog, level)(f"Executed {layer} function", {"result": result})
                return result
            except Exception as e:
                flog.error(
                    "Function failed with an exception",
                    create_error_object(
                        "INTERNAL_ERROR", f"Layer function {layer}:{function_name}", e
                    ),
                )
                raise

        return _wrapped

    def _log_wrap_async(
        self, function_name: str, func: Callable[..., Any]
    ) -> Callable[..., Any]:
        return self._log_wrap(function_name, func)

    def _log_wrap_sync(
        self, function_name: str, func: Callable[..., Any]
    ) -> Callable[..., Any]:
        return self._log_wrap(function_name, func)

    def trace(self, *a, **k):
        return self._base.trace(*a, **k)

    def debug(self, *a, **k):
        return self._base.debug(*a, **k)

    def info(self, *a, **k):
        return self._base.info(*a, **k)

    def warn(self, *a, **k):
        return self._base.warn(*a, **k)

    def error(self, *a, **k):
        return self._base.error(*a, **k)

    def apply_data(self, *a, **k):
        return self._base.apply_data(*a, **k)

    def get_id_logger(self, *a, **k):
        return self._base.get_id_logger(*a, **k)

    def get_sub_logger(self, *a, **k):
        return self._base.get_sub_logger(*a, **k)

    def get_ids(self):
        return self._base.get_ids()


def _layer_logger(
    context: CommonContext,
    sub_logger: Logger,
    layer_name: CommonLayerName | str,
    cross_layer_props: CrossLayerProps | None = None,
) -> LayerLogger:
    inner = sub_logger.get_sub_logger(str(layer_name)).apply_data(
        {"layer": str(layer_name)}
    )
    the_logger = inner.apply_data(combine_logging_props(inner, cross_layer_props))
    return _LayerLoggerImpl(context, str(layer_name), the_logger)


def _get_ids_with_runtime(
    runtime_id: str, props: Mapping[str, Any] | None
) -> list[LogId]:
    base: list[LogId] = [{"runtime_id": runtime_id}]
    if not props or "ids" not in props or not props.get("ids"):
        return base
    ids: list[LogId] = list(props.get("ids") or [])
    has_runtime = any("runtime_id" in x for x in ids)
    return ids if has_runtime else base + ids


def _get_wrap_level(
    context: CommonContext, layer_name: str, function_name: str | None
) -> str:
    getter = context.config.in_layers_core.logging.get(
        "get_function_wrap_log_level", None
    )
    level = (
        getter(layer_name, function_name)
        if callable(getter)
        else default_get_function_wrap_log_level(layer_name)
    )
    return level.value if isinstance(level, LogLevelNames) else str(level)



import inspect
import re
import unicodedata
from typing import Any

from box import Box
from pydantic import BaseModel

from .protocols import InLayersModel, ModelDefinition, ModelServices

_NON_ALNUM_RE = re.compile(r"[^a-z0-9_]+")
_UNDERSCORE_RE = re.compile(r"_+")

_META_KEY = "__in_layers_model__"


def get_model_definition(model_cls: Any) -> ModelDefinition:
    attr = getattr(model_cls, _META_KEY, None)
    if attr is None:
        raise ValueError(f"Model {model_cls} has no meta data")
    return Box(attr)


def create_model_services(models: list[InLayersModel]) -> ModelServices:
    """
    Creates a

    :param self: Description
    :param models: Description
    :type models: list[InLayersModel]
    :return: Description
    :rtype: dict[str, InLayersModel]
    """
    models = {}
    for model in models:
        meta = model.get_model_definition()
        models[meta.plural_name] = model
    return models


def is_model_class(obj: Any) -> bool:
    return (
        inspect.isclass(obj) and issubclass(obj, BaseModel) and hasattr(obj, _META_KEY)
    )


def model(*, domain: str, plural_name: str, primary_key: str = "id"):
    """
    Wrapper to turn a Pydantic model into a model capable of being used in the Simple Models system.

    :param domain: The domain this model belongs to.
    :type domain: str
    :param plural_name: The plural name of the model.
    :type plural_name: str
    """

    def decorator(cls):
        validate_plural_name(plural_name)
        existing = getattr(cls, _META_KEY, {})
        merged = {
            **existing,
            "domain": domain,
            "plural_name": plural_name,
            "primary_key": primary_key,
        }
        cls.__in_layers_model__ = merged
        return cls

    return decorator


def validate_plural_name(name: str) -> None:
    """
    Validates that the provided plural name is valid according to our rules.

    :param name: The plural name to validate.
    :type name: str
    :raises ValueError: If the name is invalid.
    """
    if not name:
        raise ValueError("Plural name cannot be empty.")
    if not re.match(r"^[A-Z][a-zA-Z0-9_]*$", name):
        raise ValueError(
            "Plural name must start with an uppercase letter and contain only alphanumeric characters and underscores."
        )


def _ascii_lower(text: str) -> str:
    """Convert text to lowercase ASCII, dropping non-ASCII characters."""
    norm = unicodedata.normalize("NFKD", text)
    ascii_bytes = norm.encode("ascii", "ignore")
    return ascii_bytes.decode("ascii").lower()


def normalize_identifier(value: str, *, max_length: int = 63) -> str:
    """
    Normalize an arbitrary string to a DB-friendly identifier:
      - ASCII only, lowercase
      - Replace non-alphanumeric with underscore
      - Collapse multiple underscores; trim leading/trailing underscores
      - Ensure it doesn't start with a digit
      - Enforce max_length with a stable hash suffix when necessary

    Default max_length 63 works for common SQL systems and is fine for Mongo collections.
    """
    if not value:
        raise ValueError("Cannot normalize empty string as identifier")
    if value[0].isdigit():
        raise ValueError(
            "Cannot normalize string that starts with a digit as identifier"
        )
    base = _ascii_lower(value)
    base = _NON_ALNUM_RE.sub("_", base)
    base = _UNDERSCORE_RE.sub("_", base).strip("_")

    if len(base) > max_length:
        raise ValueError(
            "Resulting identifier is too long. Max size is {max_length} characters."
        )

    return base



from __future__ import annotations

from collections.abc import Callable, Sequence
from dataclasses import replace
from datetime import datetime
from typing import Any, cast

from .protocols import (
    AllowableEqualitySymbols,
    BooleanQuery,
    DatastoreValueType,
    DatesAfterQuery,
    DatesBeforeQuery,
    EqualitySymbol,
    ModelSearch,
    PropertyOptions,
    PropertyQuery,
    QueryTokens,
    SortOrder,
    SortStatement,
)


def _objectize(key: str, value: Any) -> dict[str, Any]:
    return {key: value} if value is not None else {}


def is_link_token(value: Any) -> bool:
    if not isinstance(value, str):
        return False
    lower = value.lower()
    return lower in {"and", "or"}


def is_property_based_query(value: Any) -> bool:
    return isinstance(value, (PropertyQuery, DatesBeforeQuery, DatesAfterQuery))


def and_() -> BooleanQuery:
    return cast(BooleanQuery, "AND")


def or_() -> BooleanQuery:
    return cast(BooleanQuery, "OR")


def _is_date(obj: Any) -> bool:
    return isinstance(obj, datetime)


def dates_after(
    key: str,
    js_date: datetime | str,
    value_type: DatastoreValueType = DatastoreValueType.date,
    equal_to_and_after: bool = True,
) -> DatesAfterQuery:
    iso_date = js_date.isoformat() if _is_date(js_date) else cast(str, js_date)
    return DatesAfterQuery(
        type="datesAfter",
        key=key,
        date=iso_date,
        value_type=value_type,
        options=cast(Any, {"equal_to_and_after": bool(equal_to_and_after)}),
    )


def dates_before(
    key: str,
    js_date: datetime | str,
    value_type: DatastoreValueType = DatastoreValueType.date,
    equal_to_and_before: bool = True,
) -> DatesBeforeQuery:
    iso_date = js_date.isoformat() if _is_date(js_date) else cast(str, js_date)
    return DatesBeforeQuery(
        type="datesBefore",
        key=key,
        date=iso_date,
        value_type=value_type,
        options=cast(Any, {"equal_to_and_before": bool(equal_to_and_before)}),
    )


def pagination(value: Any) -> Any:
    return value


def sort(key: str, order: SortOrder = SortOrder.asc) -> SortStatement:
    if order not in (SortOrder.asc, SortOrder.dsc):
        raise ValueError("Sort must be either asc or dsc")
    return SortStatement(key=key, order=order)


def take(max_count: int | str) -> int:
    try:
        parsed = int(max_count)
    except (TypeError, ValueError):
        raise ValueError(f'Number "{max_count}" is not integerable') from None
    return parsed


def property(
    key: str,
    value: Any,
    options: PropertyOptions | None = None,
) -> PropertyQuery:
    options = options or PropertyOptions()
    equality_symbol = options.equality_symbol or EqualitySymbol.eq
    value_type = options.type or DatastoreValueType.string
    if equality_symbol not in AllowableEqualitySymbols:
        raise ValueError(f"{equality_symbol} is not a valid symbol")
    if value_type == DatastoreValueType.string and equality_symbol not in (
        EqualitySymbol.eq,
        EqualitySymbol.ne,
    ):
        raise ValueError("Cannot use a non = symbol for a string type")
    # Build options dictionary with only set values
    opts = PropertyOptions(
        case_sensitive=options.case_sensitive,
        starts_with=options.starts_with,
        ends_with=options.ends_with,
        includes=options.includes,
        # Preserve provided explicit fields; equality/type are captured above
    )
    return PropertyQuery(
        type="property",
        key=key,
        value=value,
        value_type=value_type,
        equality_symbol=equality_symbol,
        options=opts,
    )


def text_query(
    key: str,
    value: str | None,
    *,
    case_sensitive: bool | None = None,
    starts_with: bool | None = None,
    ends_with: bool | None = None,
    includes: bool | None = None,
) -> PropertyQuery:
    return property(
        key=key,
        value=value,
        options=PropertyOptions(
            case_sensitive=case_sensitive,
            starts_with=starts_with,
            ends_with=ends_with,
            includes=includes,
            type=DatastoreValueType.string,
            equality_symbol=None,
        ),
    )


def number_query(
    key: str,
    value: int | float | str | None,
    equality_symbol: EqualitySymbol = EqualitySymbol.eq,
) -> PropertyQuery:
    return property(
        key=key,
        value=value,
        options=PropertyOptions(
            type=DatastoreValueType.number,
            equality_symbol=equality_symbol,
        ),
    )


def boolean_query(key: str, value: bool | None) -> PropertyQuery:
    return property(
        key=key,
        value=value,
        options=PropertyOptions(
            type=DatastoreValueType.boolean,
            equality_symbol=EqualitySymbol.eq,
        ),
    )


def threeitize(data: Sequence[QueryTokens]) -> list[list[QueryTokens]]:
    if len(data) in (0, 1):
        return []
    if len(data) % 2 == 0:
        raise ValueError("Must be an odd number of 3 or greater.")
    three = list(data[:3])
    rest = data[2:]
    more = threeitize(rest)
    return [three, *more]


def _validate_array_or_query(obj: QueryTokens) -> None:
    if isinstance(obj, list):
        _validate_token_structure(obj)
        return
    if is_property_based_query(obj):
        return
    raise ValueError("Order of link tokens and queries invalid")


def _validate_token_types(token: QueryTokens) -> None:
    if isinstance(token, list):
        for inner in token:
            _validate_token_types(inner)
        return
    if is_property_based_query(token):
        return
    if is_link_token(token):
        return
    raise ValueError(f"Unknown token type {token}")


def _validate_token_structure(tokens: list[QueryTokens]) -> None:
    if not tokens:
        return
    first = tokens[0]
    if first in {"AND", "OR"}:
        raise ValueError("Cannot have AND or OR at the very start.")
    last = tokens[-1]
    if last in {"AND", "OR"}:
        raise ValueError("Cannot have AND or OR at the very end.")
    if all((t not in {"AND", "OR"}) for t in tokens):
        for t in tokens:
            _validate_array_or_query(t)
        return
    total_links = [t for t in tokens if t in {"AND", "OR"}]
    non_links = [t for t in tokens if t not in {"AND", "OR"}]
    if len(total_links) != len(non_links) - 1:
        raise ValueError("Must separate each statement with an AND or OR")
    threes = list(reversed(threeitize(tokens)))
    for a, link, b in threes:
        if link not in {"AND", "OR"}:
            if is_property_based_query(link):  # type: ignore[arg-type]
                raise ValueError("Must have AND/OR between property queries")
            raise ValueError("Must have AND/OR between nested queries")
        _validate_array_or_query(a)
        _validate_array_or_query(b)


def validate_model_search(search: ModelSearch) -> None:
    if not isinstance(search.query, list):
        raise ValueError("Query must be an array")
    if len(search.query) < 1:
        return
    _validate_token_types(search.query)
    _validate_token_structure(search.query)


class QueryBuilder:
    """
    Immutable-style builder for OrmSearch.
    """

    def __init__(self, data: ModelSearch | None = None):
        self._data = data or ModelSearch(query=[], take=None, sort=None, page=None)

    def _with(self, **kwargs: Any) -> QueryBuilder:
        new_data = replace(self._data, **kwargs)
        return QueryBuilder(new_data)

    def _append(self, token: QueryTokens) -> QueryBuilder:
        new_query = [*self._data.query, token]
        return self._with(query=new_query)

    # Query parts
    def property(
        self, key: str, value: Any, options: PropertyOptions | None = None
    ) -> QueryBuilder:
        return self._append(property(key, value, options))

    def dates_after(
        self,
        key: str,
        js_date: datetime | str,
        *,
        value_type: DatastoreValueType = DatastoreValueType.date,
        equal_to_and_after: bool = True,
    ) -> QueryBuilder:
        return self._append(
            dates_after(
                key=key,
                js_date=js_date,
                value_type=value_type,
                equal_to_and_after=equal_to_and_after,
            )
        )

    def dates_before(
        self,
        key: str,
        js_date: datetime | str,
        *,
        value_type: DatastoreValueType = DatastoreValueType.date,
        equal_to_and_before: bool = True,
    ) -> QueryBuilder:
        return self._append(
            dates_before(
                key=key,
                js_date=js_date,
                value_type=value_type,
                equal_to_and_before=equal_to_and_before,
            )
        )

    def complex(self, sub_builder_func: Callable[[QueryBuilder], Any]) -> QueryBuilder:
        sub_builder = QueryBuilder()
        result = sub_builder_func(sub_builder)
        if isinstance(result, QueryBuilder):
            compiled = result.compile()
            sub_tokens = compiled.query
        elif isinstance(result, ModelSearch):
            sub_tokens = result.query
        elif isinstance(result, dict) and "query" in result:
            sub_tokens = result["query"]
        else:
            raise TypeError("SubBuilder must return a QueryBuilder or ModelSearch-like")
        return self._append(cast(QueryTokens, list(sub_tokens)))

    # Linkers and non-query fields
    def and_(self) -> QueryBuilder:
        return self._append(and_())

    def or_(self) -> QueryBuilder:
        return self._append(or_())

    def take(self, count: int | str) -> QueryBuilder:
        return self._with(take=take(count))

    def sort(self, key: str, order: SortOrder = SortOrder.asc) -> QueryBuilder:
        return self._with(sort=sort(key, order))

    def pagination(self, value: Any) -> QueryBuilder:
        return self._with(page=pagination(value))

    def compile(self) -> ModelSearch:
        # Return a snapshot ModelSearch
        return ModelSearch(
            query=list(self._data.query),
            take=self._data.take,
            sort=self._data.sort,
            page=self._data.page,
        )


def query_builder() -> QueryBuilder:
    return QueryBuilder()



from __future__ import annotations

import json
from collections.abc import Mapping
from typing import Any

from box import Box

from ..libs import (
    combine_cross_layer_props,
    is_cross_layer_props,
    is_object_shaped_cross_layer_props,
    normalize_cross_layer_props,
)
from ..protocols import CrossLayerProps, Logger, LogLevelNames


def default_get_function_wrap_log_level(layer_name: str) -> LogLevelNames:
    if layer_name in ("features", "entries"):
        return LogLevelNames.info
    if layer_name == "services":
        return LogLevelNames.trace
    return LogLevelNames.debug


def combine_logging_props(
    logger: Logger, cross_layer_props: CrossLayerProps | None = None
) -> Mapping[str, Any]:

    base: CrossLayerProps = {"logging": {"ids": logger.get_ids()}}
    final = combine_cross_layer_props(base, cross_layer_props or {})  # type: ignore[arg-type]
    return final["logging"]


def cap_for_logging(input: Any, max_size: int = 50000) -> Any:
    def safe_stringify(obj: Any) -> str:
        try:
            return json.dumps(obj, default=str)
        except Exception:
            return "[Unserializable]"

    if not isinstance(input, (list, dict)):
        return input
    if len(safe_stringify(input)) <= max_size:
        return input
    if isinstance(input, list):
        build: list[Any] = []
        for item in input:
            candidate = [*build, item]
            if (
                len(
                    safe_stringify(
                        [*candidate, f"[truncated, original length: {len(input)}]"]
                    )
                )
                > max_size
            ):
                return [*build, f"[truncated, original length: {len(input)}]"]
            build.append(item)
        return build
    keys = list(input.keys())
    out: dict[str, Any] = {}
    for k in keys:
        out[k] = input[k]
        truncated = dict(out)
        truncated["[truncated]"] = f"original keys: {len(keys)}"
        if len(safe_stringify(truncated)) > max_size:
            return truncated
    return out


_CROSS_PARAM_NAMES = {
    "crossLayer",
    "cross_layer",
    "crossLayerProps",
    "cross_layer_props",
}


def extract_cross_layer_props(
    args: list[Any],
    kwargs: Mapping[str, Any] | None = None,
) -> tuple[list[Any], dict[str, Any], Box | None]:
    """
    Extract cross-layer props from the last positional argument if present,
    otherwise look for a named kwarg using any supported cross param name.
    Always converts to Box so framework and user functions receive Box.
    Returns (args_without_cross, kwargs_without_cross, cross_as_box_or_none).
    """
    args_no_cross = list(args) if args else []
    kwargs_copy: dict[str, Any] = dict(kwargs or {})
    cross_layer_props: CrossLayerProps | None = None
    # 1) Check last positional
    if args_no_cross:
        last = args_no_cross[-1]
        if is_cross_layer_props(last):  # type: ignore[arg-type]
            cross_layer_props = last  # type: ignore[assignment]
            args_no_cross = args_no_cross[:-1]
            return (
                args_no_cross,
                kwargs_copy,
                normalize_cross_layer_props(cross_layer_props),
            )
        if is_object_shaped_cross_layer_props(last):  # type: ignore[arg-type]
            cross_layer_props = last  # type: ignore[assignment]
            args_no_cross = args_no_cross[:-1]
            return (
                args_no_cross,
                kwargs_copy,
                normalize_cross_layer_props(cross_layer_props),
            )
    # 2) Check kwargs for known names
    for name in _CROSS_PARAM_NAMES:
        if name in kwargs_copy:
            cross = kwargs_copy[name]
            if is_cross_layer_props(cross) or is_object_shaped_cross_layer_props(cross):
                cross_layer_props = cross  # type: ignore[assignment]
            kwargs_copy.pop(name)  # type: ignore[assignment]
    return args_no_cross, kwargs_copy, normalize_cross_layer_props(cross_layer_props)



from __future__ import annotations

from typing import Any

from box import Box

from ..libs import is_config, validate_config
from ..protocols import CommonContext, CoreNamespace, FeaturesContext

globals_name = CoreNamespace.globals.value


class GlobalsFeatures:
    def __init__(self, context: FeaturesContext):
        self.context = context

    def load_globals(self, environment_or_config: Any) -> CommonContext:
        services = self.context.services[globals_name]
        if not services:
            raise RuntimeError(f"Services for {globals_name} not found")
        config = (
            environment_or_config
            if is_config(environment_or_config)
            else services.load_config()
        )
        validate_config(config)
        common_globals: CommonContext = Box(
            config=config,
            root_logger=services.get_root_logger(),
            constants=services.get_constants(),
        )
        return common_globals


def create(context: FeaturesContext) -> GlobalsFeatures:
    return GlobalsFeatures(context)



from __future__ import annotations

from collections.abc import Mapping
from types import ModuleType
from typing import Any

from ..models.libs import is_model_class
from ..protocols import Domain, LayerContext, ServicesContext


def _iter_models_from_module(module: ModuleType) -> dict[str, Any]:
    result: dict[str, Any] = {}
    for name in dir(module):
        try:
            attr = getattr(module, name)
        except Exception:  # noqa: S112
            continue
        if is_model_class(attr):
            result[name] = attr
    return result


def _iter_models_from_mapping(mapping: Mapping[str, Any]) -> dict[str, Any]:
    result: dict[str, Any] = {}
    for k, v in mapping.items():
        if is_model_class(v):
            result[str(k)] = v
    return result


def _iter_models_from_object(obj: Any) -> dict[str, Any]:
    result: dict[str, Any] = {}
    for name in dir(obj):
        if name.startswith("_"):
            continue
        try:
            attr = getattr(obj, name)
        except Exception:  # noqa: S112
            continue
        if is_model_class(attr):
            result[name] = attr
    return result


def _iter_model_candidates(container: Any) -> Mapping[str, Any]:
    """
    Discover model classes from a container. Supports modules, mappings, and objects.
    """
    if container is None:
        return {}
    if isinstance(container, ModuleType):
        return _iter_models_from_module(container)
    if isinstance(container, Mapping):
        return _iter_models_from_mapping(container)  # type: ignore[arg-type]
    return _iter_models_from_object(container)


def _build_domain_model_index(
    context: ServicesContext,
) -> Mapping[str, Mapping[str, Any]]:
    domains = getattr(context.config.in_layers_core, "domains", []) or []
    domain_name_to_models: dict[str, dict[str, Any]] = {}
    for domain in domains:
        domain_name = getattr(domain, "name", None)
        if not domain_name:
            continue
        models_attr = getattr(domain, "models", None)
        domain_name_to_models[domain_name] = dict(_iter_model_candidates(models_attr))
    return domain_name_to_models


class _ModelResolver:
    def __init__(self, index: Mapping[str, Mapping[str, Any]]):
        self.__index = index

    def get_model(self, namespace: str, model_name: str) -> Any:
        models_for_namespace = self.__index.get(namespace)
        if not models_for_namespace:
            raise KeyError(f"No models found for namespace '{namespace}'")
        model_cls = models_for_namespace.get(model_name)
        if not model_cls:
            raise KeyError(f"Model '{model_name}' not found in namespace '{namespace}'")
        return model_cls


class LayersServices:
    def get_model_props(self, context: ServicesContext):
        """
        Provide model props needed by domains to construct models.
        For the Python port, we expose only `context` and a `get_model` resolver.
        """
        resolver = _ModelResolver(_build_domain_model_index(context))

        return {
            "context": context,
            "get_model": resolver.get_model,
        }

    def load_layer(self, app: Domain, layer: str, context: LayerContext):
        layer_instance = getattr(app, layer, None)
        if not layer_instance or not hasattr(layer_instance, "create"):
            return None
        instance = layer_instance.create(context)
        if instance is None:
            raise RuntimeError(
                f"App {app.get('name')} did not return an instance layer {layer}"
            )
        return instance


def create() -> LayersServices:
    return LayersServices()



from __future__ import annotations

from datetime import UTC, datetime
from typing import Any
from uuid import uuid4

from box import Box

from .protocols import (
    AND,
    OR,
    BackendProtocol,
    BooleanQuery,
    DatastoreValueType,
    DatesAfterQuery,
    DatesBeforeQuery,
    EqualitySymbol,
    ModelDefinition,
    PropertyOptions,
    PropertyQuery,
    Query,
    QueryTokens,
    SortOrder,
    SortStatement,
)


class MemoryBackend(BackendProtocol):
    @staticmethod
    def create_unique_connection_string() -> str:
        return f"memory://{uuid4().hex}"

    def __init__(self) -> None:
        # Storage layout:
        #   key -> f"{domain}_{plural_name}"
        #   value -> { primary_key_value -> data_mapping }
        self.__buckets: dict[str, dict[Any, dict[str, Any]]] = {}
        # Auto-increment sequences per bucket (when PK not provided)
        self.__sequences: dict[str, int] = {}

    def __bucket_key(self, model) -> str:
        meta = model.get_model_definition()
        return f"{meta.domain}_{meta.plural_name}"

    def __get_bucket(self, model) -> dict[Any, dict[str, Any]]:
        key = self.__bucket_key(model)
        if key not in self.__buckets:
            self.__buckets[key] = {}
            self.__sequences[key] = 0
        return self.__buckets[key]

    def get_backend_name(self) -> str:
        return "memory"

    def get_raw_client(self) -> Any:
        return self

    def create(self, model, data):  # type: ignore[override]
        bucket = self.__get_bucket(model)
        payload = dict(data)

        pk_name = model.get_primary_key_name()
        pk_value = payload.get(pk_name)
        if pk_value is None:
            # allocate a simple auto-increment integer id
            bucket_key = self.__bucket_key(model)
            self.__sequences[bucket_key] = self.__sequences.get(bucket_key, 0) + 1
            pk_value = self.__sequences[bucket_key]
            payload[pk_name] = pk_value

        bucket[pk_value] = payload
        return dict(payload)

    def retrieve(self, model, id):  # type: ignore[override]
        bucket = self.__get_bucket(model)
        record = bucket.get(id)
        if record is None:
            return None
        return dict(record)

    def update(self, model, id, data):  # type: ignore[override]
        bucket = self.__get_bucket(model)
        existing = bucket.get(id)
        if existing is None:
            raise KeyError(f"Instance with id {id!r} not found")

        updated = dict(existing)
        updated.update(dict(data))
        # Ensure primary key field remains consistent with the id argument
        pk_name = model.get_primary_key_name()
        updated[pk_name] = id

        bucket[id] = updated
        return dict(updated)

    def delete(self, model, id):  # type: ignore[override]
        bucket = self.__get_bucket(model)
        bucket.pop(id, None)

    def search(self, model, query):
        bucket = self.__get_bucket(model)
        records = list(bucket.values())
        filtered = [r for r in records if _matches_query_tokens(r, query.query)]
        sorted_records = _apply_sort(filtered, query.sort)
        limited = _apply_take(sorted_records, query.take)
        return Box(instances=[dict(x) for x in limited], page=query.page)

    def bulk_insert(self, model, data):
        for item in data:
            self.create(model, item)

    def bulk_delete(self, model, ids):
        for id in ids:
            self.delete(model, id)

    def dispose(self) -> None:
        self.__buckets.clear()
        self.__sequences.clear()


class DefaultModelFactory:
    """
    Core fallback model factory. Returns a no-op backend.
    Domains can provide their own factory services to supply real backends.
    """

    def __init__(self, context: Any):
        self.__context = context

    def get_model_backend(self, model_definition: ModelDefinition):  # noqa: ARG002
        return MemoryBackend()


def create(context: Any) -> DefaultModelFactory:
    return DefaultModelFactory(context)


# ================
# Search utilities
# ================


def _apply_sort(records: list[dict[str, Any]], sort_stmt: SortStatement | None):
    if not sort_stmt:
        return records
    key = sort_stmt.key
    reverse = sort_stmt.order == SortOrder.dsc

    def _safe_key(r: dict[str, Any]):
        v = r.get(key)
        return (v is None, str(v))

    return sorted(records, key=_safe_key, reverse=reverse)


def _apply_take(records: list[dict[str, Any]], take_value: int | None):
    if not take_value:
        return records
    return records[: int(take_value)]


def _matches_query_tokens(record: dict[str, Any], tokens: list[QueryTokens]) -> bool:
    if not tokens:
        return True
    if _has_links(tokens):
        return _evaluate_linked_tokens(record, tokens)
    return all(_evaluate_token(record, t) for t in tokens)


def _has_links(tokens: list[QueryTokens]) -> bool:
    return any(t == "AND" or t == "OR" for t in tokens)  # noqa: PLR1714


def _evaluate_token(record: dict[str, Any], token: QueryTokens) -> bool:
    if isinstance(token, list):
        return _matches_query_tokens(record, token)
    if (
        token == AND or token == OR  # noqa: PLR1714
    ):  # pragma: no cover - structure validated elsewhere
        return True
    return _evaluate_query(record, token)


def _evaluate_linked_tokens(record: dict[str, Any], tokens: list[QueryTokens]) -> bool:
    threes = _threeitize(tokens)
    if not threes:
        # Single element with link tokens present shouldn't happen, fallback to AND across items
        return all(_evaluate_token(record, t) for t in tokens)
    # Evaluate left-to-right using provided links
    result: bool | None = None
    for a, link, b in threes:
        left = _evaluate_token(record, a)
        right = _evaluate_token(record, b)
        current = _combine_link(left, link, right)
        result = current if result is None else _combine_link(result, link, current)
    return bool(result)


def _combine_link(a: bool, link: BooleanQuery, b: bool) -> bool:
    if link == "AND":
        return a and b
    return a or b


def _threeitize(
    data: list[QueryTokens],
) -> list[tuple[QueryTokens, BooleanQuery, QueryTokens]]:
    if len(data) in (0, 1):
        return []
    if len(data) % 2 == 0:
        raise ValueError("Must be an odd number of 3 or greater.")
    three = (data[0], _as_link(data[1]), data[2])
    rest = data[2:]
    more = _threeitize(rest)
    return [three, *more]


def _as_link(value: QueryTokens) -> BooleanQuery:
    if value in {"AND", "OR"}:
        return value
    raise ValueError("Must have AND/OR between statements")


def _evaluate_query(record: dict[str, Any], q: Query) -> bool:
    if isinstance(q, PropertyQuery):
        return _match_property(record, q)
    if isinstance(q, DatesBeforeQuery):
        return _match_dates_before(record, q)
    if isinstance(q, DatesAfterQuery):
        return _match_dates_after(record, q)
    return False


def _match_property(record: dict[str, Any], q: PropertyQuery) -> bool:
    actual = record.get(q.key, None)
    opts = q.options
    if q.value_type == DatastoreValueType.string:
        return _match_string(actual, q.value, q.equality_symbol, opts)
    if q.value_type == DatastoreValueType.number:
        return _match_number(actual, q.value, q.equality_symbol)
    if q.value_type == DatastoreValueType.boolean:
        return _match_boolean(actual, q.value, q.equality_symbol)
    # Fallback to equality for other types
    return _apply_equality(actual, q.value, q.equality_symbol)


def _match_string(
    actual: Any, expected: Any, symbol: EqualitySymbol, opts: PropertyOptions
) -> bool:
    a = _to_str(actual)
    e = _to_str(expected)
    # Default to case-insensitive matching unless explicitly set
    case_sensitive = opts.case_sensitive if opts.case_sensitive is not None else False
    if not case_sensitive:
        a = a.lower() if a is not None else a
        e = e.lower() if e is not None else e
    # Pattern options
    if opts.starts_with:
        return _apply_symbol_bool(
            a.startswith(e) if a is not None and e is not None else False, symbol
        )
    if opts.ends_with:
        return _apply_symbol_bool(
            a.endswith(e) if a is not None and e is not None else False, symbol
        )
    if opts.includes:
        return _apply_symbol_bool(
            (e in a) if a is not None and e is not None else False, symbol
        )
    # Default equals / not equals
    return _apply_equality(a, e, symbol)


def _apply_symbol_bool(condition: bool, symbol: EqualitySymbol) -> bool:
    if symbol == EqualitySymbol.ne:
        return not condition
    return condition


def _match_number(actual: Any, expected: Any, symbol: EqualitySymbol) -> bool:
    try:
        a = float(actual)
        e = float(expected)
    except (TypeError, ValueError):
        return False
    return _compare(a, e, symbol)


def _match_boolean(actual: Any, expected: Any, symbol: EqualitySymbol) -> bool:
    if symbol not in (EqualitySymbol.eq, EqualitySymbol.ne):
        return False
    a = bool(actual) if actual is not None else None
    e = bool(expected) if expected is not None else None
    return _apply_equality(a, e, symbol)


def _apply_equality(a: Any, b: Any, symbol: EqualitySymbol) -> bool:
    if symbol == EqualitySymbol.eq:
        return a == b
    if symbol == EqualitySymbol.ne:
        return a != b
    # For non-equality symbols, attempt numeric comparison when possible
    try:
        return _compare(float(a), float(b), symbol)
    except (TypeError, ValueError):
        return False


def _compare(a: float, b: float, symbol: EqualitySymbol) -> bool:  # noqa: PLR0911
    if symbol == EqualitySymbol.lt:
        return a < b
    if symbol == EqualitySymbol.lte:
        return a <= b
    if symbol == EqualitySymbol.gt:
        return a > b
    if symbol == EqualitySymbol.gte:
        return a >= b
    if symbol == EqualitySymbol.eq:
        return a == b
    if symbol == EqualitySymbol.ne:
        return a != b
    return False


def _match_dates_after(record: dict[str, Any], q: DatesAfterQuery) -> bool:
    actual = record.get(q.key, None)
    if q.value_type == DatastoreValueType.date:
        a = _to_datetime(actual)
        b = _to_datetime(q.date)
        if a is None or b is None:
            return False
        if q.options.equal_to_and_after:
            return a >= b
        return a > b
    # String comparison
    a_str = _to_str(actual)
    b_str = _to_str(q.date)
    if a_str is None or b_str is None:
        return False
    if q.options.equal_to_and_after:
        return a_str >= b_str
    return a_str > b_str


def _match_dates_before(record: dict[str, Any], q: DatesBeforeQuery) -> bool:
    actual = record.get(q.key, None)
    if q.value_type == DatastoreValueType.date:
        a = _to_datetime(actual)
        b = _to_datetime(q.date)
        if a is None or b is None:
            return False
        if q.options.equal_to_and_before:
            return a <= b
        return a < b
    a_str = _to_str(actual)
    b_str = _to_str(q.date)
    if a_str is None or b_str is None:
        return False
    if q.options.equal_to_and_before:
        return a_str <= b_str
    return a_str < b_str


def _to_str(value: Any) -> str | None:
    if value is None:
        return None
    return str(value)


def _to_datetime(value: Any):  # noqa: PLR0911

    if value is None:
        return None
    if isinstance(value, datetime):
        return value
    if isinstance(value, (int, float)):
        try:
            return datetime.fromtimestamp(value, tz=UTC)
        except Exception:
            return None
    if isinstance(value, str):
        try:
            # ISO-8601
            return datetime.fromisoformat(value.replace("Z", "+00:00"))
        except Exception:
            return None
    return None



from __future__ import annotations

import functools
import inspect
from collections.abc import Mapping
from typing import Any

from box import Box

from ..globals.libs import extract_cross_layer_props
from ..libs import (
    combine_cross_layer_props,
    get_layers_unavailable,
    normalize_cross_layer_props,
)
from ..models.libs import get_model_definition, is_model_class
from ..models.protocols import InLayersModel, PrimaryKeyType
from ..models.services import create_in_layers_model
from ..protocols import (
    CommonContext,
    CoreNamespace,
    FeaturesContext,
)
from ..utils import rgetattr

_CROSS_PARAM_NAMES = {
    "crossLayer",
    "cross_layer",
    "crossLayerProps",
    "cross_layer_props",
}

_INTERNAL_MODELS_KEY = "__in_layers_models"


class _CrudsWrapper:
    def __init__(self, in_layers_model: InLayersModel):
        self._im = in_layers_model

    def get_model(self):
        return self._im

    def create(self, data=None, **kwargs):
        inst = self._im.create(data, **kwargs)
        return inst.to_pydantic()

    def retrieve(self, id):
        inst = self._im.retrieve(id)
        return None if inst is None else inst.to_pydantic(Box(no_validation=True))

    def update(self, id, **kwargs):
        inst = self._im.update(id, **kwargs)
        return inst.to_pydantic()

    def delete(self, id):
        self._im.delete(id)

    def search(self, query):
        res = self._im.search(query)
        instances = [i.to_pydantic(Box(no_validation=True)) for i in res.instances]
        page = getattr(res, "page", None)
        return Box(instances=instances, page=page)

    def bulk_insert(self, data: list[Mapping]) -> list[Mapping]:
        return self._im.bulk_insert(data)

    def bulk_delete(self, ids: list[PrimaryKeyType]) -> None:
        self._im.bulk_delete(ids)


class _FeatureCruds:
    def __init__(self, base_crud):
        self._base = base_crud

    def get_model(self):
        return self._base.get_model()

    def create(self, data=None, **kwargs):
        return self._base.create(data, **kwargs)

    def retrieve(self, id):
        return self._base.retrieve(id)

    def update(self, id, **kwargs):
        return self._base.update(id, **kwargs)

    def delete(self, id):
        return self._base.delete(id)

    def search(self, query):
        return self._base.search(query)

    def bulk_insert(self, data: list[Mapping]) -> list[Mapping]:
        return self._base.bulk_insert(data)

    def bulk_delete(self, ids: list[PrimaryKeyType]) -> None:
        return self._base.bulk_delete(ids)


def _resolve_backend_for_model(
    features: LayersFeatures, layer_context: Mapping[str, Any], model_cls: Any
):
    backend = _resolve_backend_from_domain_provider(features, layer_context, model_cls)
    if backend is None:
        backend = _resolve_backend_from_core(features, model_cls)
    return backend


def _resolve_backend_from_domain_provider(
    features: LayersFeatures, layer_context: Mapping[str, Any], model_cls: Any
):
    backend_domain = rgetattr(
        features, "context.config.in_layers_core.models.model_backend", None
    )
    if not backend_domain:
        return None
    services_map = layer_context.get("services", None)
    if not services_map:
        raise ValueError("Services map is not found in layer context.")
    provider = services_map.get(backend_domain)
    if not provider:
        raise ValueError(f"Provider {backend_domain} does not have a services layer.")
    return provider.get_model_backend(get_model_definition(model_cls))


def _resolve_backend_from_core(features: LayersFeatures, model_cls: Any):
    core_factory = features.context.services.get(CoreNamespace.models.value)  # type: ignore[index]
    if not core_factory:
        raise ValueError(f"Model {model_cls} has no core factory")
    return core_factory.get_model_backend(get_model_definition(model_cls))


def _build_in_layers_models_for_app(
    features: LayersFeatures,
    layer_context_local: Mapping[str, Any],
    discovered_local: Mapping[str, Any],
) -> Mapping[str, Any]:
    wrapped: dict[str, Any] = {}
    for name, model_cls in discovered_local.items():
        backend = _resolve_backend_for_model(
            features, Box(layer_context_local), model_cls
        )
        if backend is None:
            try:
                meta = get_model_definition(model_cls)
                plural = Box(meta).plural_name if meta else name
                wrapped[plural] = model_cls
            except Exception:
                wrapped[name] = model_cls
            continue
        simple = create_in_layers_model(model_cls, backend)
        try:
            plural = Box(simple.get_model_definition()).plural_name
        except Exception:
            try:
                meta = get_model_definition(model_cls)
                plural = Box(meta).plural_name if meta else name
            except Exception:
                plural = name
        wrapped[plural] = simple
    return Box(wrapped)


def _build_services_cruds(simple_models_map: Mapping[str, Any]) -> Mapping[str, Any]:
    cruds_wrappers: dict[str, Any] = {}
    for plural_name, simple_model in dict(simple_models_map).items():
        cruds_wrappers[plural_name] = _CrudsWrapper(simple_model)
    return Box(cruds_wrappers)


def _maybe_add_services_cruds(
    features: LayersFeatures,
    layer_context: Mapping[str, Any],
    app: Mapping[str, Any],
    final_layer: Mapping[str, Any],
) -> Mapping[str, Any]:
    services_cruds_flag = bool(
        rgetattr(
            features, "context.config.in_layers_core.models.model_services_cruds", False
        )
    )
    features_cruds_flag = bool(
        rgetattr(
            features, "context.config.in_layers_core.models.model_features_cruds", False
        )
    )
    services_cruds_enabled = services_cruds_flag or features_cruds_flag
    if not services_cruds_enabled:
        return final_layer
    models_ctx = layer_context.get("models", {})  # type: ignore[assignment]
    domain_models_entry = (
        models_ctx.get(app.name, {}) if isinstance(models_ctx, Mapping) else {}
    )
    simple_models = (
        domain_models_entry.get(_INTERNAL_MODELS_KEY)
        if isinstance(domain_models_entry, Mapping)
        else Box({})
    )
    cruds_wrappers = _build_services_cruds(simple_models or Box({}))
    new_final = dict(final_layer)
    new_final["cruds"] = cruds_wrappers
    return new_final


def _maybe_add_features_cruds(
    features: LayersFeatures,
    layer_context: Mapping[str, Any],
    app: Mapping[str, Any],
    final_layer: Mapping[str, Any],
) -> Mapping[str, Any]:
    if not _is_features_cruds_enabled(features):
        return final_layer
    service_cruds = _get_service_cruds_for_domain(layer_context, app)
    if not isinstance(service_cruds, Mapping):
        return final_layer
    feature_wrappers = _wrap_service_cruds_for_features(service_cruds)
    new_final = dict(final_layer)
    new_final["cruds"] = Box(feature_wrappers)
    return new_final


def _is_features_cruds_enabled(features: LayersFeatures) -> bool:
    try:
        return bool(features.context.config.in_layers_core.models.model_features_cruds)
    except Exception:
        return False


def _get_service_cruds_for_domain(
    layer_context: Mapping[str, Any], app: Mapping[str, Any]
) -> Mapping[str, Any] | None:
    services_map = layer_context.get("services", {})  # type: ignore[assignment]
    service_for_domain = (
        services_map.get(app.name) if isinstance(services_map, Mapping) else None
    )
    service_cruds = (
        service_for_domain.get("cruds")
        if isinstance(service_for_domain, Mapping)
        else None
    )
    return service_cruds if isinstance(service_cruds, Mapping) else None


def _wrap_service_cruds_for_features(
    service_cruds: Mapping[str, Any],
) -> Mapping[str, Any]:
    feature_wrappers: dict[str, Any] = {}
    for plural_name, svc_crud in dict(service_cruds).items():
        feature_wrappers[plural_name] = _FeatureCruds(svc_crud)
    return feature_wrappers


def _create_wrapper_with_metadata(original_func: Any, inner_callable: Any) -> Any:
    """
    Return a new wrapper with original_func's metadata/signature, adding
    an optional cross_layer_props parameter when not explicitly present.
    """
    wrapped = functools.wraps(original_func)(inner_callable)
    try:
        sig = getattr(original_func, "__signature__", inspect.signature(original_func))
        params = list(sig.parameters.values())
        has_cross = any(p.name in _CROSS_PARAM_NAMES for p in params)
        if not has_cross:
            new_param = inspect.Parameter(
                "cross_layer_props",
                kind=inspect.Parameter.POSITIONAL_OR_KEYWORD,
                default=None,
            )
            wrapped.__signature__ = sig.replace(parameters=[*params, new_param])
        else:
            wrapped.__signature__ = sig
        wrapped.__wrapped__ = getattr(original_func, "__wrapped__", original_func)
    except Exception:  # noqa: S110
        pass
    return wrapped


def _iter_properties_for_wrap(obj: Any):
    if isinstance(obj, Mapping):
        for k, v in obj.items():
            yield from (k, v)
        return
    for name in dir(obj):
        if name.startswith("_"):
            continue
        try:
            attr = getattr(obj, name)
        except Exception:  # noqa: S112
            continue
        yield name, attr


def _get_params_for_func(fn: Any) -> list[inspect.Parameter]:
    return list(inspect.signature(fn).parameters.values())


def _get_explicit_positional_params(
    params: list[inspect.Parameter],
) -> list[inspect.Parameter]:
    return [
        p
        for p in params
        if p.kind
        in (inspect.Parameter.POSITIONAL_ONLY, inspect.Parameter.POSITIONAL_OR_KEYWORD)
    ]


def _has_var_positional(params: list[inspect.Parameter]) -> bool:
    return any(p.kind is inspect.Parameter.VAR_POSITIONAL for p in params)


def _trim_surplus_args_for_params(
    pos_args: list[Any], params: list[inspect.Parameter]
) -> list[Any]:
    if _has_var_positional(params):
        return pos_args
    explicit = _get_explicit_positional_params(params)
    surplus = len(pos_args) - len(explicit)
    return pos_args[surplus:] if surplus > 0 else pos_args


def _find_cross_kw_name_in_params(params: list[inspect.Parameter]) -> str | None:
    for p in params:
        if p.name in _CROSS_PARAM_NAMES:
            return p.name
    return None


def _call_with_optional_cross(
    f,
    args_no_cross: list[Any],
    kwargs_no_cross: dict[str, Any],
    cross_layer_props: Mapping[str, Any] | None,
) -> Any:
    params = _get_params_for_func(f)
    args_no_cross = _trim_surplus_args_for_params(args_no_cross, params)
    if cross_layer_props is None:
        return f(*args_no_cross, **kwargs_no_cross)

    cross_kw = _find_cross_kw_name_in_params(params)
    if cross_kw:
        if cross_kw not in kwargs_no_cross:
            kwargs_no_cross = dict(kwargs_no_cross)
            kwargs_no_cross[cross_kw] = cross_layer_props
        return f(*args_no_cross, **kwargs_no_cross)

    # No cross-like kw param; avoid positional injection if *args is present
    if _has_var_positional(params):
        return f(*args_no_cross, **kwargs_no_cross)

    explicit = _get_explicit_positional_params(params)
    if len(args_no_cross) + 1 == len(explicit):
        # Check if the parameter that would receive cross_layer_props is already in kwargs
        # to avoid "multiple values for argument" error
        next_param_index = len(args_no_cross)
        if next_param_index < len(explicit):
            next_param_name = explicit[next_param_index].name
            if next_param_name in kwargs_no_cross:
                # Can't inject positionally - would conflict with keyword arg
                return f(*args_no_cross, **kwargs_no_cross)
        return f(*args_no_cross, cross_layer_props, **kwargs_no_cross)
    return f(*args_no_cross, **kwargs_no_cross)


def _make_passthrough_for_log(f):
    def _inner(log, *args, **kwargs):  # noqa: ARG001
        return f(*args, **kwargs)

    return _create_wrapper_with_metadata(f, _inner)


def _should_copy_direct_layer_key(key: str) -> bool:
    return key in (
        "_logging",
        "root_logger",
        "log",
        "constants",
        "config",
        "models",
        "get_models",
        "cruds",
    )


def _wrap_domain_mapping_for_load(
    features: LayersFeatures,
    domain_value: Mapping[str, Any],
    logger_ids: Any,
) -> Mapping[str, Any]:
    domain_data: dict[str, Any] = {}
    for property_name, func in domain_value.items():
        if not callable(func):
            domain_data[property_name] = func
            continue
        wrapped_func = features._make_wrapped(func, logger_ids)
        domain_data[property_name] = wrapped_func
    return domain_data


def _build_wrapped_context_for_load(
    features: LayersFeatures,
    ctx: Mapping[str, Any],
    logger_ids: Any,
) -> Mapping[str, Any]:
    wrapped: dict[str, Any] = {}
    for layer_key, layer_data in ctx.items():
        if _should_copy_direct_layer_key(layer_key) or not isinstance(
            layer_data, Mapping
        ):
            wrapped[layer_key] = layer_data
            continue
        final_layer_data: dict[str, Any] = {}
        for domain_key, domain_value in layer_data.items():
            if not isinstance(domain_value, Mapping):
                final_layer_data[domain_key] = domain_value
                continue
            final_layer_data[domain_key] = _wrap_domain_mapping_for_load(
                features,
                domain_value,
                logger_ids,
            )
        wrapped[layer_key] = final_layer_data
    return wrapped


class LayersFeatures:
    def __init__(self, context: FeaturesContext):
        self.context = context

    def _get_layer_context(
        self, common_context: Mapping[str, Any], layer: Mapping[str, Any] | None
    ) -> Mapping[str, Any]:
        if layer:
            merged = Box(common_context)
            return merged + Box(layer)
        return common_context

    def _make_wrapped(self, f, logger_ids):
        def _inner2(*args, **kwargs):
            args_no_cross, kwargs_no_cross, cross_layer_props = (
                extract_cross_layer_props(list(args), dict(kwargs))
            )
            # Normalize to Box here so we never pass a Pydantic/object instance into combine or the user's function
            cross_as_box = normalize_cross_layer_props(cross_layer_props)
            base = {"logging": {"ids": logger_ids}}
            combined = combine_cross_layer_props(base, cross_as_box or {})  # type: ignore[arg-type]
            return _call_with_optional_cross(
                f, args_no_cross, kwargs_no_cross, combined
            )

        return _create_wrapper_with_metadata(f, _inner2)

    def _wrap_layer_functions(
        self,
        loaded_layer: Any,
        layer_logger,
        app_name: str,
        layer: str,
        ignore_layer_functions: list[str],
    ):
        out: dict[str, Any] = {}
        logger_ids = layer_logger.get_ids()
        for property_name, func in _iter_properties_for_wrap(loaded_layer):
            if not callable(func):
                out[property_name] = func
                continue
            function_level_key = f"{app_name}.{layer}.{property_name}"
            # Always wrap for cross-layer props
            cross_wrapped = self._make_wrapped(func, logger_ids)
            # Only add logging wrapper when not ignored
            if _should_ignore_path(ignore_layer_functions, function_level_key):
                wrapped = cross_wrapped
            else:
                logged_func = layer_logger._log_wrap(
                    property_name, _make_passthrough_for_log(cross_wrapped)
                )
                wrapped = _create_wrapper_with_metadata(func, logged_func)
            out[property_name] = wrapped
        return out

    def _inject_models_context(
        self, app: Mapping[str, Any], layer_context: Mapping[str, Any]
    ) -> Mapping[str, Any]:
        """
        When in the services layer, attach model discovery helpers and
        precomputed simple model wrappers for the app into the context.
        """
        models_attr = getattr(app, "models", None)
        if models_attr is None:
            return layer_context
        discovered = _iter_model_candidates(models_attr)
        simple_models_box = _build_in_layers_models_for_app(
            self, layer_context, discovered
        )

        def get_models() -> Mapping[str, Any]:
            return simple_models_box

        models_context = dict(layer_context.get("models", {}))
        models_context[app.name] = {
            **models_context.get(app.name, {}),
            "get_models": get_models,
            _INTERNAL_MODELS_KEY: simple_models_box,  # available for internal use
        }
        new_ctx = dict(layer_context)
        new_ctx["models"] = models_context
        return new_ctx

    def _load_composite_layer(
        self,
        app: Mapping[str, Any],
        composite_layers,
        common_context: Mapping[str, Any],
        previous_layer: Mapping[str, Any] | None,  # noqa: ARG002
        anti_layers_fn,  # noqa: ARG002
    ):
        result = {}
        for layer in composite_layers:
            layer_logger = (
                self.context.root_logger.get_logger(
                    Box(
                        common_context,
                    )
                )
                .get_app_logger(app.name)
                .get_layer_logger(layer)
            )
            the_context = dict(common_context)
            the_context["log"] = layer_logger
            wrapped_context = the_context
            loaded = self.context.services[CoreNamespace.layers.value].load_layer(
                app,
                layer,
                Box(
                    wrapped_context,
                ),
            )
            if loaded:
                ignore_layer_functions = self.context.config.in_layers_core.logging.get(
                    "ignore_layer_functions", []
                )
                final_layer = self._wrap_layer_functions(
                    loaded, layer_logger, app.name, layer, ignore_layer_functions
                )
                result = {**result, layer: {app.name: final_layer}}
        return result

    def _load_layer(
        self,
        app: Mapping[str, Any],
        current_layer: str,
        common_context: Mapping[str, Any],
        previous_layer: Mapping[str, Any] | None,
    ):
        layer_context1 = self._get_layer_context(common_context, previous_layer)
        layer_logger = (
            self.context.root_logger.get_logger(Box(layer_context1))
            .get_app_logger(app.name)
            .get_layer_logger(current_layer)
        )
        layer_context = dict(layer_context1)
        layer_context["log"] = layer_logger

        # If this is the services layer, attach model discovery helpers
        if str(current_layer) == "services":
            layer_context = self._inject_models_context(app, layer_context)

        logger_ids = layer_logger.get_ids()
        ignore_layer_functions = self.context.config.in_layers_core.logging.get(
            "ignore_layer_functions", []
        )
        wrapped_context = _build_wrapped_context_for_load(
            self, layer_context, logger_ids
        )

        loaded = self.context.services.in_layers_core_layers.load_layer(
            app,
            current_layer,
            Box(wrapped_context),
        )
        if not loaded:
            return {}
        final_layer = self._wrap_layer_functions(
            loaded, layer_logger, app.name, current_layer, ignore_layer_functions
        )
        # Inject model CRUD wrappers into services/features when enabled
        if str(current_layer) == "services":
            final_layer = _maybe_add_services_cruds(
                self, layer_context, app, final_layer
            )
        if str(current_layer) == "features":
            final_layer = _maybe_add_features_cruds(
                self, layer_context, app, final_layer
            )
        return {current_layer: {app.name: final_layer}}

    def load_layers(self):
        layers_in_order = self.context.config.in_layers_core.layer_order
        anti_layers = get_layers_unavailable(layers_in_order)
        core_layers_to_ignore = [
            f"services.{CoreNamespace.layers.value}",
            f"services.{CoreNamespace.globals.value}",
            f"features.{CoreNamespace.layers.value}",
            f"features.{CoreNamespace.globals.value}",
        ]
        starting_context: CommonContext = {k: v for k, v in self.context.items() if k not in core_layers_to_ignore}  # type: ignore[return-value]
        apps = self.context.config.in_layers_core.domains
        existing_layers = starting_context
        for app in apps:
            previous_layer = {}
            for layer in layers_in_order:
                if isinstance(layer, list):
                    layer_instance = self._load_composite_layer(
                        app,
                        layer,
                        {k: v for k, v in existing_layers.items() if k != "log"},
                        previous_layer,
                        anti_layers,
                    )
                else:
                    layer_instance = self._load_layer(
                        app,
                        layer,
                        {k: v for k, v in existing_layers.items() if k != "log"},
                        previous_layer,
                    )
                if not layer_instance:
                    previous_layer = {}
                    continue
                # Deep-merge by layer so we accumulate domains instead of overwriting
                new_context = dict(existing_layers)
                for layer_key, layer_value in layer_instance.items():
                    if (
                        layer_key in new_context
                        and isinstance(new_context[layer_key], Mapping)
                        and isinstance(layer_value, Mapping)
                    ):
                        merged_layer = dict(new_context[layer_key])
                        merged_layer.update(layer_value)
                        new_context[layer_key] = merged_layer
                    else:
                        new_context[layer_key] = layer_value
                if "log" in new_context:
                    new_context = {k: v for k, v in new_context.items() if k != "log"}
                existing_layers = new_context
                previous_layer = layer_instance
        return Box(
            existing_layers,
        )


def create(context: FeaturesContext) -> LayersFeatures:
    return LayersFeatures(context)


def _should_ignore_path(ignore_list: list[str], dotted: str) -> bool:
    if not ignore_list:
        return False
    dotted = dotted.strip().strip(".")
    for pattern in ignore_list:
        if not pattern:
            continue
        pat = str(pattern).strip().strip(".")
        if not pat:
            continue
        if dotted == pat or dotted.startswith(f"{pat}."):
            return True
    return False


def _iter_model_candidates(container: Any) -> dict[str, Any]:
    """
    Discover model classes from a container by inspecting either mapping items
    or public attributes. Returns a name->class dictionary.
    """
    result: dict[str, Any] = {}
    if container is None:
        return result
    # Mapping support
    if isinstance(container, Mapping):
        for k, v in container.items():
            if is_model_class(v):
                result[str(k)] = v
        return result
    # Attribute-based discovery for modules/objects
    for name in dir(container):
        if name.startswith("_"):
            continue
        try:
            attr = getattr(container, name)
        except Exception:  # noqa: S112
            continue
        if is_model_class(attr):
            result[name] = attr
    return result



from collections.abc import Mapping
from typing import Any, Self

from box import Box
from pydantic import BaseModel

from .backends import MemoryBackend
from .libs import get_model_definition
from .protocols import (
    BackendProtocol,
    InLayersModel,
    InLayersModelInstance,
    ModelDefinition,
    ModelSearchResult,
    PrimaryKeyType,
    ToPydanticOptions,
)


class _InLayersModelInstanceImpl(InLayersModelInstance):

    def __init__(self, model: InLayersModel, data: Mapping):
        self.__model = model
        self.__data = data

    class _GetterAccessor:
        def __init__(self, data: Mapping[str, Any]):
            self.__data = data

        def __getattr__(self, name: str):
            def _getter():
                try:
                    return self.__data[name]
                except Exception:
                    return None

            return _getter

        def __dir__(self):
            try:
                return list(set(object.__dir__(self) + list(self.__data.keys())))
            except Exception:
                return object.__dir__(self)

    def get_model(self) -> InLayersModel:
        return self.__model

    def to_dict(self) -> Box:
        return Box(self.__data)

    def to_pydantic(self, options: ToPydanticOptions | None = None) -> BaseModel:
        return self.__model.to_pydantic(self.__data, options)

    def validate(self) -> None:
        self.__model.validate(self.__data)

    def delete(self) -> None:
        primary_key = self.get_primary_key()
        self.__model.delete(primary_key)

    def update(self, **kwargs) -> Self:
        primary_key = self.get_primary_key()
        return self.__model.update(primary_key, **kwargs)

    def get_primary_key(self) -> PrimaryKeyType:
        return self.__model.get_primary_key(self.__data)

    @property
    def get(self):
        """
        Provides zero-arg getters for each property in the instance data:
            instance.get.id()
            instance.get.name()
        """
        # Expose a dynamic accessor bound to the current data mapping
        return _InLayersModelInstanceImpl._GetterAccessor(self.__data)  # type: ignore[arg-type]


class _InLayersModelImpl(InLayersModel):

    def __init__(self, pydantic_model: BaseModel, backend: BackendProtocol):
        if get_model_definition(pydantic_model) is None:
            raise ValueError("Model is not decorated with @model")
        self.__model = pydantic_model
        self.__backend = backend

    def instance(
        self, data: Mapping | None = None, **kwargs: Any
    ) -> InLayersModelInstance:
        """
        Wraps the data to give model capabilities.
        This does NOT save the data.

        :param self: Description
        :param data: Optional mapping to initialize
        :type data: Mapping | None
        """
        base: dict[str, Any] = {}
        if isinstance(data, Mapping):
            base.update(dict(data))
        if kwargs:
            base.update(kwargs)
        return _InLayersModelInstanceImpl(self, base)

    def get_model_definition(self) -> ModelDefinition:
        return get_model_definition(self.__model)

    def validate(self, data: Mapping) -> None:
        self.__model.validate(data)

    def create(
        self, data: Mapping | None = None, **kwargs: Any
    ) -> InLayersModelInstance:
        payload: dict[str, Any] = {}
        if isinstance(data, Mapping):
            payload.update(dict(data))
        if kwargs:
            payload.update(kwargs)
        created_data = self.__backend.create(self, payload)
        return self.instance(created_data)

    def retrieve(self, id) -> InLayersModelInstance | None:
        data = self.__backend.retrieve(self, id)
        if data is None:
            return None
        return self.instance(data)

    def update(self, id, **kwargs) -> InLayersModelInstance:
        new_data = self.__backend.update(self, id, kwargs)
        return self.instance(new_data)

    def delete(self, id) -> None:
        self.__backend.delete(self, id)

    def search(self, query) -> ModelSearchResult:
        result = self.__backend.search(self, query)
        final_container = Box(
            instances=[self.instance(item) for item in result.instances],
            page=result.page,
        )
        return final_container

    def bulk_insert(self, data: list[Mapping]) -> None:
        self.__backend.bulk_insert(self, data)

    def bulk_delete(self, ids: list[PrimaryKeyType]) -> None:
        self.__backend.bulk_delete(self, ids)

    def get_primary_key_name(self) -> str:
        meta = get_model_definition(self.__model)
        return meta.primary_key

    def get_primary_key(self, model_data: Mapping) -> PrimaryKeyType:
        pk_name = self.get_primary_key_name()
        return model_data[pk_name]

    def to_pydantic(
        self, data: Mapping, options: ToPydanticOptions | None = None
    ) -> BaseModel:
        if options and options.no_validation:
            # Use model_construct to skip validation
            return self.__model.model_construct(**data)
        return self.__model(**data)


def create_in_layers_model(
    pydantic_model: BaseModel, backend: BackendProtocol
) -> InLayersModel:
    """
    Public helper to wrap a Pydantic model into an InLayersModel with the provided backend.
    """
    return _InLayersModelImpl(pydantic_model, backend)


class ModelsServices:

    def __init__(self, context):
        self.__context = context

    def get_model_backend(
        self, model_definition: ModelDefinition  # noqa: ARG002
    ) -> BackendProtocol:
        return MemoryBackend()


def create(context) -> ModelsServices:
    return ModelsServices(context)



from __future__ import annotations

from collections.abc import Mapping
from enum import Enum
from typing import Any, Literal, Protocol, Self, TypeAlias

from box import Box
from pydantic import BaseModel
from pydantic.dataclasses import dataclass

# ==============================
# Search/Query Type Definitions
# ==============================


class EqualitySymbol(Enum):
    eq = "="
    lt = "<"
    lte = "<="
    gt = ">"
    gte = ">="
    ne = "!="


AllowableEqualitySymbols: tuple[str, ...] = (
    EqualitySymbol.eq,
    EqualitySymbol.lt,
    EqualitySymbol.lte,
    EqualitySymbol.gt,
    EqualitySymbol.gte,
    EqualitySymbol.ne,
)


class DatastoreValueType(Enum):
    string = "string"
    number = "number"
    date = "date"
    object = "object"
    boolean = "boolean"


class SortOrder(Enum):
    asc = "asc"
    dsc = "dsc"


@dataclass(frozen=True)
class SortStatement:
    key: str
    order: SortOrder


@dataclass(frozen=True)
class PropertyOptions:
    case_sensitive: bool | None = None
    starts_with: bool | None = None
    ends_with: bool | None = None
    includes: bool | None = None
    type: DatastoreValueType | None = None
    equality_symbol: EqualitySymbol | None = None


class ToPydanticOptions(Protocol):
    no_validation: bool | None


@dataclass(frozen=True)
class PropertyQuery:
    type: Literal["property"]
    key: str
    value: Any
    value_type: DatastoreValueType
    equality_symbol: EqualitySymbol
    options: PropertyOptions


@dataclass(frozen=True)
class _DatesAfterOptions:
    equal_to_and_after: bool


@dataclass(frozen=True)
class DatesAfterQuery:
    type: Literal["datesAfter"]
    key: str
    date: str
    value_type: DatastoreValueType
    options: _DatesAfterOptions


@dataclass(frozen=True)
class _DatesBeforeOptions:
    equal_to_and_before: bool


@dataclass(frozen=True)
class DatesBeforeQuery:
    type: Literal["datesBefore"]
    key: str
    date: str
    value_type: DatastoreValueType
    options: _DatesBeforeOptions


# Queries supported
type Query = PropertyQuery | DatesAfterQuery | DatesBeforeQuery

AND = Literal["AND"]
OR = Literal["OR"]

# Boolean link tokens
BooleanQuery: TypeAlias = Literal["AND", "OR"]  # noqa: UP040

# Recursive token structure:
# - either a nested list of tokens
# - or a boolean link
# - or a concrete query
type QueryTokens = "list[QueryTokens]" | BooleanQuery | Query


@dataclass(frozen=True)
class ModelSearch:
    """
    A recursive, structured search object used by backends.
    """

    query: list[QueryTokens]
    take: int | None = None
    sort: SortStatement | None = None
    page: Any | None = None


class ModelSearchResult(Protocol):
    instances: list[Mapping]
    page: Any | None


type PrimaryKeyType = str | int


@dataclass(frozen=True)
class ModelDefinition(Protocol):
    domain: str
    plural_name: str
    primary_key: str


class _ZeroArgGetter(Protocol):
    def __call__(self) -> Any: ...


class _GetterAccessor(Protocol):
    """
    Dynamic accessor that exposes zero-arg getters for instance properties.
    Example: instance.get.id() -> Any
    """

    def __getattr__(self, name: str) -> _ZeroArgGetter: ...


class InLayersModelInstance(Protocol):
    """
    An instance of a simple model. Wraps the data itself and provides methods to work with it.
    """

    #: Gets the underlying model
    def get_model(self) -> InLayersModel: ...

    #: Updates the instance with the provided data
    def update(self, **kwargs) -> Self: ...

    #: Deletes the instance from the backend
    def delete(self) -> None: ...

    #: Converts the instance data to a Box dictionary
    def to_dict(self) -> Box: ...

    #: Converts the instance data to a Pydantic model
    def to_pydantic(self, options: ToPydanticOptions | None = None) -> BaseModel: ...

    #: Validates the instance data against the model schema
    def validate(self) -> None: ...

    #: Gets the primary key value of the instance
    def get_primary_key(self) -> PrimaryKeyType: ...

    @property
    def get(self) -> _GetterAccessor: ...


class InLayersModel(Protocol):
    """
    A Simple Model that describes (abstractly) data that is stored in a backend.
    Also provides CRUDS interface for working with the data.
    """

    #: Gets metadata about the model
    def get_model_definition(self) -> ModelDefinition: ...

    #: Creates a non-persisted instance wrapper from mapping or keyword args
    def instance(
        self, data: Mapping | None = None, **kwargs: Any
    ) -> InLayersModelInstance: ...

    #: Creates a new instance of the model with the provided data
    def create(
        self, data: Mapping | None = None, **kwargs: Any
    ) -> InLayersModelInstance: ...

    #: Retrieves an instance of the model by its primary key
    def retrieve(self, id: PrimaryKeyType) -> InLayersModelInstance | None: ...

    #: Validates the provided data against the model schema
    def validate(self, data: Mapping) -> None: ...

    #: Updates an existing instance of the model by its primary key with the provided data
    def update(self, id: PrimaryKeyType, **kwargs: Any) -> InLayersModelInstance: ...

    #: Deletes an instance of the model by its primary key
    def delete(self, id: PrimaryKeyType) -> None: ...

    #: Searches for instances of the model matching the query
    def search(self, query: ModelSearch) -> ModelSearchResult: ...

    #: Bulk inserts a list of instances of the model with the provided data
    def bulk_insert(self, data: list[Mapping]) -> None: ...

    #: Bulk deletes a list of instances of the model by their primary keys
    def bulk_delete(self, ids: list[PrimaryKeyType]) -> None: ...

    #: Gets the name of the primary key field
    def get_primary_key_name(self) -> str: ...

    #: Gets the primary key value from the provided model data
    def get_primary_key(self, model_data: Mapping) -> PrimaryKeyType: ...

    #: Converts the model data to a Pydantic model
    def to_pydantic(
        self, data: Mapping, options: ToPydanticOptions | None = None
    ) -> BaseModel: ...


class BackendProtocol(Protocol):

    def create(self, model: InLayersModel, data: Mapping) -> Mapping: ...
    def retrieve(self, model: InLayersModel, id: PrimaryKeyType) -> Mapping | None: ...
    def update(
        self, model: InLayersModel, id: PrimaryKeyType, data: Mapping
    ) -> Mapping: ...
    def delete(self, model: InLayersModel, id: PrimaryKeyType) -> None: ...
    def search(self, model: InLayersModel, query: ModelSearch) -> ModelSearchResult: ...
    def bulk_insert(self, model: InLayersModel, data: list[Mapping]) -> None: ...
    def bulk_delete(self, model: InLayersModel, ids: list[PrimaryKeyType]) -> None: ...
    def dispose(self) -> None: ...
    def get_raw_client(self) -> Any: ...
    def get_backend_name(self) -> str: ...


ModelServices = Mapping[str, InLayersModel]
